/*  tmap_scanline_per - Pentium-optimized assembly version		*/
/*  written by Brian Raiter, Feb 1998.					*/
   
/*  This version of tmap_scanline_per avoids the unshrinkable integer	*/
/*  divides of the normal version by foisting those calculations off	*/
/*  onto the FPU.							*/
   
/*  The function needs to calculate trunc(u/z) and trunc(v/z) at every	*/
/*  iteration. During the init it loads u, v, z, and du, dv, dz into	*/
/*  the FP stack, and then calculates 1/z. It then multiplies this	*/
/*  reciprocal with u and v respectively, and truncates the products	*/
/*  to integers as it writes them out.					*/
   
/*  Of course, a FP divide is hardly any faster than idiv. But: the	*/
/*  division operates in parallel, so we can get everything else done	*/
/*  while that's going. In fact, the FPU is hardly idle from the moment	*/
/*  we start. Putting everything possible onto the FPU stack also frees	*/
/*  us from having to store those values in registers or local memory.	*/
   
/*  (These darling little Pentium optimizations will suck mud on a 486.	*/
/*  You will not be pleased. Needless to say, the results will be even	*/
/*  less enjoyable on a 486SX. And finally, this function should be	*/
/*  ripped out at once if you're considering making use of any MMX	*/
/*  capabilities. Consider yourself warned.)				*/


.globl	c_tmap_scanline_per

.extern	fx_xleft, fx_xright, fx_y, bytes_per_row
.extern	fx_u, fx_v, fx_z, fx_l, fx_du_dx, fx_dv_dx, fx_dz_dx, fx_dl_dx
.extern	pixptr, gr_fade_table, write_buffer
.extern Transparency_on

.local	ubyz
.local	vbyz
.local	ctlwd
.local	ctlwdx


	/* Our local data storage. */

.data

	.balign	4
ubyz:	.long	0
vbyz:	.long	0
ctlwd:	.word	0
ctlwdx:	.word	0


	/* The function itself. */

.text

	.balign	4

c_tmap_scanline_per:

	/* Save registers that the compiler might be using. */

	pushl	%ebp
	pushl	%edi
	pushl	%esi

	/* Set the FPU's rounding to truncate (chop mode).		*/
	/* Also, change the precision to the lowest possible setting	*/
	/* (still enough for our needs, and makes for a faster FDIV).	*/

	fnstcw	(ctlwd)
	movw	(ctlwd), %ax
	orb	$0x0C, %ah
	andb	$0xFC, %ah
	movw	%ax, (ctlwdx)
	fldcw	(ctlwdx)

	/* Get u, v, z, dudx, dvdx, dzdx into the FP stack.		*/
	/* In parallel with this, get our loop counter into ebp.	*/
	/* ebp = fx_xright - fx_xleft + 1				*/
	/* Also get our base source and destination pointers.		*/
	/* Then start calculating 1/z.					*/

	fildl	(fx_dz_dx)
	movl	(fx_dv_dx), %eax
	movl	(fx_v), %ebx
	fildl	(fx_du_dx)
	sall	$6, %eax
	movl	%eax, (vbyz)
	fildl	(vbyz)
	sall	$6, %ebx
	movl	%ebx, (vbyz)
	fildl	(fx_z)
	movl	(write_buffer), %edi
	movl	(pixptr), %esi
	fildl	(fx_u)
	movl	(fx_xright), %ebp
	movl	(fx_xleft), %ecx
	fildl	(vbyz)
	subl	%ecx, %ebp
	incl	%ebp
	fld1
	jle	Leave_now		/* ebp <= 0 means nothing to do */
	fdiv	%st(3), %st

	/* edi = write_buffer + fx_xleft + (bytes_per_row * fx_y)	*/
	/* Get l and dldx into edx and ecx.				*/

	movl	(bytes_per_row), %eax
	addl	%ecx, %edi
	mull	(fx_y)
	addl	%eax, %edi
	movl	(fx_l), %edx
	movl	(fx_dl_dx), %ecx
	sarl	$8, %edx
	sarl	$8, %ecx

	/* Which loop do we use? */

	testl	$0xFFFFFFFF, (Transparency_on)
	jnz	Loop_trans_on

	/* Transparency_on is zero: use simpler loop. */

	.balign	4

Loop_trans_off:

	/* FP stack: 1/z, v, u, z, dv, du, dz */

	/* Calculate v/z and u/z, and retrieve as integers.		*/
	/* Increment u, v, and z by dudx, dvdx, and dzdx respectively.	*/
	/* Then begin calculating the new 1/z.				*/

	fld	%st(0)
	fmul	%st(2), %st	/* st = v * (1 / z) */
	fxch	%st(5)
	fadd	%st, %st(2)	/* v += dvdx */
	fxch	%st(5)
	fxch	%st(1)
	fmul	%st(3), %st	/* st = u * (1 / z) */
	fxch	%st(6)
	fadd	%st, %st(3)	/* u += dudx */
	fxch	%st(6)
	fistpl	(ubyz)
	fxch	%st(6)
	fadd	%st, %st(3)	/* z += dzdx */
	fxch	%st(6)
	fistpl	(vbyz)
	fld1
	fdiv	%st(3), %st	/* st = 1 / z */

	/* al = pixptr(((v/z) & 0x0FC0) | ((u/z) & 0x003F)) */

	movl	(vbyz), %eax
	movl	(ubyz), %ebx
	andl	$0x0FC0, %eax
	andl	$0x003F, %ebx
	orl	%eax, %ebx
	xorl	%eax, %eax
	movb	(%ebx,%esi), %al

	/* *dest++ = gr_fade_table((l & 0xFF00) | al)	*/
	/* l += dldx					*/

	movl	%edx, %ebx
	andl	$0xFF00, %ebx
	orl	%ebx, %eax
	addl	%ecx, %edx
	movb	gr_fade_table(%eax), %al
	movb	%al, (%edi)
	incl	%edi

	/* Decrement and loop. */

	decl	%ebp
	jnz	Loop_trans_off
	jmp	Leave_now

	.balign	4

	/* Second loop reiteration point */

Leave_pixel_be:
	addl	%ecx, %edx
	incl	%edi
	decl	%ebp
	jz	Leave_now

	.balign	4

	/* Transparency_on is nonzero: use this loop. */

Loop_trans_on:

	/* FP stack: 1/z, v, u, z, dv, du, dz */

	/* Calculate v/z and u/z, and retrieve as integers.		*/
	/* Increment u, v, and z by dudx, dvdx, and dzdx respectively.	*/
	/* Then begin calculating the new 1/z.				*/

	fld	%st(0)
	fmul	%st(2), %st	/* st = v * (1 / z) */
	fxch	%st(5)
	fadd	%st, %st(2)	/* v += dvdx */
	fxch	%st(5)
	fxch	%st(1)
	fmul	%st(3), %st	/* st = u * (1 / z) */
	fxch	%st(6)
	fadd	%st, %st(3)	/* u += dudx */
	fxch	%st(6)
	fistpl	(ubyz)
	fxch	%st(6)
	fadd	%st, %st(3)	/* z += dzdx */
	fxch	%st(6)
	fistpl	(vbyz)
	fld1
	fdiv	%st(3), %st	/* st = 1 / z */

	/* al = pixptr(((v/z) & 0x0FC0) | ((u/z) & 0x003F)) */

	movl	(vbyz), %eax
	movl	(ubyz), %ebx
	andl	$0x0FC0, %eax
	andl	$0x003F, %ebx
	orl	%eax, %ebx
	xorl	%eax, %eax
	movb	(%ebx,%esi), %al

	/* Write to dest only if pixptr value is not transparent. */

	cmpb	$0x0FF, %al
	je	Leave_pixel_be
	movl	%edx, %ebx
	andl	$0xFF00, %ebx
	orl	%ebx, %eax
	addl	%ecx, %edx
	movb	gr_fade_table(%eax), %al
	movb	%al, (%edi)
	incl	%edi

	/* Decrement and loop. */

	decl	%ebp
	jnz	Loop_trans_on

	/* We're out of here. */

Leave_now:
	
	ffree	%st(6)
	ffree	%st(5)
	ffree	%st(4)
	ffree	%st(3)
	ffree	%st(2)
	ffree	%st(1)
	ffree	%st
	fldcw	(ctlwd)
	popl	%esi
	popl	%edi
	popl	%ebp
	ret
